*****************************************************************************
* FILENAME: TMSEC.S
* CONTENTS: Declaration of secondaries used by TM.
* AUTHOR  : John Olwoch
* DATE    :
*****************************************************************************

*----------------------------------------------------------------------------
* NAME: Errmsg
* FUNCTION: Display error messages.
* STACK OPERATION: ( %n --> )
* CALLS: none
* CALLED BY:
*----------------------------------------------------------------------------
'
::
  '
  ::
    DISPROW7
    VERYVERYSLOW                        ( *wait 3 secs* )
    #25 #19 BLANKIT                     ( *clear message area* )
  ;

  { LAM err
    LAM Show_cause
  } BIND

  ERRBEEP

  ::
    "Error:" DISPROW6

    LAM err %IP %1 %= case              ( *run error* )
      ::
	::
	  LAM err %FP .1 %= case
	    "TM in halt state"

	  LAM err %FP .2 %= case
	    "Not in start position"
	;
	LAM Show_cause EVAL
      ;

    LAM err %IP %2 %= case              ( *view error* )
      ::
	"End of tape"
	LAM Show_cause EVAL
      ;

    LAM err %IP %3 %= case              ( *command error* )
      ::
	"."  ' LAM move_direction STO   ( *prepare to halt machine* )
	::
	  LAM err %FP .1 %= case
	    ::
	      "Next state invalid"
	    ;

	  LAM err %FP .2 %= case
	    ::
	      "Scan direction invalid"
	    ;

	  LAM err %FP .3 %= case
	    "Unknown state"
	;
	LAM Show_cause EVAL

	LAM err 3.3 %= ITE              ( *show faulty command or state )
	  ::
	    "State = " LAM this_state &$
	  ;
	  ::
	    "Command = " LAM command &$
	  ;
	  DISPROW7
      ;
  ;

  ABND
;

*----------------------------------------------------------------------------
* NAME: Validinput
* FUNCTION: Check that the input data (TAPE and TMPRG) is valid.
* STACK OPERATION: ( --> )
* CALLS: none
* CALLED BY: Setup
*----------------------------------------------------------------------------
'
::
  "TAPE" $>ID SAFE@_HERE ITE

  ::
    ' LAM tape STO

    "PRG" $>ID SAFE@_HERE ITE
      ::
	LENCOMP ' LAM states STO        ( *save number of states* )
      ;
      ::
	ERRBEEP
	"Error:" NEWLINE$ &$
	"PRG is missing" &$
	DISPSTATUS2
	VERYVERYSLOW
	TRUE ' LAM exit STO
      ;
  ;

  ::
    ERRBEEP
    "Error:" NEWLINE$ &$
    "TAPE is missing" &$
    DISPSTATUS2
    VERYVERYSLOW
    TRUE ' LAM exit STO
  ;

;

*----------------------------------------------------------------------------
* NAME: Frame
* FUNCTION: Frame the display.
* STACK OPERATION: ( --> )
* CALLS: none
* CALLED BY: Setup
*----------------------------------------------------------------------------
'
::
    0  0 130  0 LINEON
  130  0 130 36 LINEON
    0 36 130 36 LINEON
    0  0   0 36 LINEON
;

*----------------------------------------------------------------------------
* NAME: Find_last_page
* FUNCTION: Find last page number.
* STACK OPERATION: ( --> )
* CALLS:
* CALLED BY: Setup
*----------------------------------------------------------------------------
'
::
  LAM tape LEN$ UNCOERCE                ( *find length of tape* )
  %13 %/ DUP %FP %0 %= ITE              ( *get no of sections for display* )
    %IP                                 ( *exact no* )
    ::
      %IP %1 %+                         ( *increment if not exact no* )
    ;
  COERCE ' LAM last_page STO            ( *no of display pages* )
;

*----------------------------------------------------------------------------
* NAME: Left_page_indicator
* FUNCTION: Display/delete the left page indicator.
* STACK OPERATION: ( #n --> )
* CALLS: none
* CALLED BY:
*----------------------------------------------------------------------------
'
::
  1 #=ITE

  ::                                    ( *display indicator* )
    LAM left_indicator
    HARDBUFF 6 1 GROB!
    32 SetUserFlag                      ( *set left page flag* )
  ;

  ::                                    ( *delete indicator* )
    HARDBUFF 6 1 15 10 GROB!ZERODRP     ( grob x1 y1 x2 y2 --> )
    32 ClrUserFlag
  ;

;

*----------------------------------------------------------------------------
* NAME: Right_page_indicator
* FUNCTION: Display/delete the right page indicator.
* STACK OPERATION: ( #n --> )
* CALLS: none
* CALLED BY:
*----------------------------------------------------------------------------
'
::
  1 #=ITE

  ::                                    ( *display indicator* )
    LAM right_indicator
    HARDBUFF 116 1 GROB!
    33 SetUserFlag                      ( *set right page flag* )
  ;

  ::                                    ( *delete indicator* )
    HARDBUFF 116 1 125 10
    GROB!ZERODRP                        ( grob x1 y1 x2 y2 --> )
    33 ClrUserFlag
  ;

;

*----------------------------------------------------------------------------
* NAME: Symbol
* FUNCTION: Display/delete a symbol at the current scanner position.
* STACK OPERATION: ( #d #p #s #e --> )
* CALLS: none
* CALLED BY:
*----------------------------------------------------------------------------
'
::
  { 
    LAM display                         ( *display/delete* )
    LAM posn                            ( *position of symbol on tape* )
					( *if delete then dummy, = 0* )
    LAM start                           ( *start display/delete here* )
    LAM end                             ( *stop display/delete here* )
  } BIND

  LAM display 1 #=ITE

  ::                                    ( *display symbol* )
    LAM end 1 #+ LAM start
    DO
      LAM tape LAM posn
      INDEX@ LAM start #- #+ DUP
      SUB$ DUP "0" EQUAL                ( *read symbol and check if "0"* )
      35 TestUserFlag NOT AND ITE       ( *check "0" display flag* )
	DROP                            ( *discard if 35 is FALSE* )
	::
	  $>BIGGROB                     ( *convert to 5x9 char* )
	  HARDBUFF
	  LAM symbol_col INDEX@
	  NTHCOMPDROP 14 GROB!          ( *display symbol* )
	;
    LOOP
  ;

  ::                                    ( *delete symbol* )
    LAM end 1 #+ LAM start
    DO
      HARDBUFF                          ( --> grob )
      LAM symbol_col INDEX@
      NTHCOMDDUP 6 #+ 14 SWAP 23        ( x1 x2 y1 y2 --> )
      GROB!ZERODRP                      ( grob x1 y1 x2 y2 --> )
    LOOP
  ;

  ABND
;

*----------------------------------------------------------------------------
* NAME: Scanner
* FUNCTION: Display/delete scanner.
* STACK OPERATION: ( #n --> )
* CALLS: none
* CALLED BY:
*----------------------------------------------------------------------------
'
::
  1 #=ITE

  ::                                    ( *display scanner* )
    LAM scanner 
    HARDBUFF 
    LAM scanner_col
    LAM scanner_pos 
    NTHCOMPDROP 27 GROB!                ( *display scanner in new position* )
  ;

  ::                                    ( *delete scanner* )
    HARDBUFF                            ( --> grob )
    LAM scanner_col 
    LAM scanner_pos
    NTHCOMDDUP 9 #+ 27 SWAP 36          ( x1 x2 y1 y2 --> )
    GROB!ZERODRP                        ( grob x1 y1 x2 y2 --> )
  ;

;

*----------------------------------------------------------------------------
* NAME: New_page
* FUNCTION: Display prev/next section of tape.
* STACK OPERATION: ( #n --> )
* CALLS: Left_page_indicator, Right_page_indicator, Symbol
* CALLED BY: Move
*----------------------------------------------------------------------------
'
::
  { LAM d } BIND                        ( *-1 = page left, 1 = page right * )

* Display next page.

  LAM d 1 #= ITE                        ( *left or right page ?* )

  ::                                    ( *if right* )
    LAM this_page 1 #+ 
    ' LAM this_page STO                 ( *then increment page no* )
    0 0 1 13 LAM Symbol EVAL            ( *delete symbols on current page* )
    LAM this_page LAM last_page #=      ( *is new page the last page ?* )
    33 TestUserFlag AND IT              ( *right page indicator on ?* )
      ::                                ( *if both true* )
	0 LAM Right_page_indicator      ( *then delete right page indicator* )
	EVAL
      ;
    32 TestUserFlag ?SKIP               ( *left page indicator off ?* )
      ::                                ( *if off* )
	1 LAM Left_page_indicator       ( *then display it* )
	EVAL
      ;
    1 LAM position
    1 13 LAM Symbol EVAL                ( *display new page* )
  ;

  ::                                    ( *if left* )
    LAM this_page 1 #- 
    ' LAM this_page STO                 ( *then decrement page no* )
    0 0 1 13 LAM Symbol EVAL            ( *delete symbols on current page* )
    LAM this_page 1 #=                  ( *is new page the first page ?* )
    32 TestUserFlag AND IT              ( *left page indicator on ?* )
      ::                                ( *if both true* )
	0 LAM Left_page_indicator       ( *then delete left page indicator* )
	EVAL
      ;
    LAM last_page 1 #>                  ( *more then one page ?* )
    33 TestUserFlag NOT AND IT          ( *right page indicator off ?* )
      ::                                ( *if both true* )
	1 LAM Right_page_indicator      ( *then display right page ind*)
	EVAL
      ;
    1 LAM position 12 #-                ( *from beginning of prev page* )
    1 13 LAM Symbol EVAL                ( *display new page* )
  ;

* Display/delete scanner if viewing only.

  34 TestUserFlag IT                    ( *viewing* )
    ::
      LAM scanner_page LAM this_page
      #= ITE 1 0 LAM Scanner EVAL       ( *display/delete scanner* )
      34 ClrUserFlag                    ( *reset flag* )
    ;

  ABND
;

*----------------------------------------------------------------------------
* NAME: Move
* FUNCTION: Move scanner.
* STACK OPERATION: ( #n --> )
* CALLS: New_page, Scanner
* CALLED BY: main
*----------------------------------------------------------------------------
'
::
  { LAM d } BIND

  0 LAM Scanner EVAL                    ( *del scanner from old position* )

  LAM d 1 #=ITE                         ( *which direction ?* )

  ::                                    ( *right* )
    LAM scanner_pos 13 #=ITE
      ::
	1 ' LAM scanner_pos STO         ( *move to start of next page* )
	LAM d LAM New_page EVAL
      ;
      ::
	LAM scanner_pos LAM d #+
	' LAM scanner_pos STO
      ;
  ;

  ::                                    ( *left* )
    LAM scanner_pos 1 #=ITE
      ::
	13 ' LAM scanner_pos STO        ( *move to end of prev page* )
	LAM d LAM New_page EVAL
      ;
      ::
	LAM scanner_pos LAM d #+
	' LAM scanner_pos STO           ( *move scanner one step left* )
      ;
  ;

  1 LAM Scanner EVAL                    ( *display scanner in new position* )

  ABND
;

*----------------------------------------------------------------------------
* NAME: Nextstate
* FUNCTION: Get the next state.
* STACK OPERATION: ( --> )
* CALLS:
* CALLED BY: T_machine
*----------------------------------------------------------------------------
'
::
  LAM command 5 6 SUB$                  ( *find next state* )
  ' LAM next_state STO                  ( *save it* )

  LAM next_state palparse DROP         ( *assume it can be parsed*)
  TYPEREAL? ?SKIP                      ( *check it is REAL* )
    ::
      3.1 LAM Errmsg EVAL               ( *invalid state* )
    ;
;

*----------------------------------------------------------------------------
* NAME: Direction
* FUNCTION: Find next direction to move, after executing command.
*           L = Move left, R = Move right, N = No movement.
* STACK OPERATION: ( --> )
* CALLS: Move
* CALLED BY: main
*----------------------------------------------------------------------------
'
::
  LAM command 8 8 SUB$ DUP              ( *extract next direction* )
  ' LAM move_direction STO              ( *save it* )

  "."   EQUAL ?SKIP                     ( *halt ?* )

  ::                                    ( *else get ready to move* )
    LAM move_direction "L" EQUALcase

    ::                                  ( *left* )
      LAM position #1 #= IT             ( *reached end of tape ?* )

      ::                                ( *if so then...* )
	LAM more_tape LAM tape &$
	' LAM tape STO                  ( *add more tape at left end* )
	LAM last_page 1 #+
	' LAM last_page STO             ( *one more page* )
	32 TestUserFlag ?SKIP
	  ::                            ( *indicate left page available* )
	    1 LAM Left_page_indicator
	    EVAL
	  ;
	LAM this_page 1 #+ 
	' LAM this_page STO             ( *current page is now page 2* )
	14 ' LAM position STO           ( *point to first loc of page 2* )
      ;

      LAM position 1 #-                 ( *decrement symbol pointer* )
      ' LAM position STO                ( *save it* )
      MINUSONE LAM Move EVAL            ( *move left* )
    ;

    LAM move_direction "R" EQUALcase

    ::                                  ( *right* )
      LAM position
      LAM tape LEN$ #= IT               ( *reached end of tape ?* )

      ::                                ( *if so then...* )
	LAM tape LAM more_tape &$
	' LAM tape STO                  ( *add more tape at right end* )
	LAM last_page 1 #+
	' LAM last_page STO             ( *one more page* )
	35 TestUserFlag IT              ( *displaying "0"?* )
	  ::                            ( *if so, display new tape section* )
	    1 LAM position 1 #+
	    LAM scanner_pos 1 #+ 13     ( *from next scanner posn to end* )
	    LAM Symbol EVAL
	  ;
	33 TestUserFlag ?SKIP
	  ::                            ( *indicate right page available* )
	    1 LAM Right_page_indicator
	    EVAL
	  ;
      ;

      LAM position 1 #+                 ( *increment symbol pointer* )
      ' LAM position STO                ( *save it* )
      1 LAM Move EVAL                   ( *move right* )
    ;

    LAM move_direction "N" EQUALNOTcase

    ::                                  ( *do nothing if 'N'* )
      3.2 LAM Errmsg EVAL               ( *else display error message* )
    ;
  ;

;

*----------------------------------------------------------------------------
* NAME: Write
* FUNCTION: Write symbol to tape if not already there.
* STACK OPERATION: ( --> )
* CALLS: Symbol
* CALLED BY: T_machine
*----------------------------------------------------------------------------
'
::
  LAM command 7 7 SUB$                  ( *extract next symbol to write* )
  { LAM symbol } BIND

  LAM tape LAM position DUP SUB$
  LAM symbol EQUAL ?SKIP                ( *if symbol not already on tape* )

  ::                                    ( *write it on the tape* )
    0 0 LAM scanner_pos DUP
    LAM Symbol EVAL                     ( *delete old symbol* )

    LAM tape 1 LAM position
    1 #- SUB$                           ( *extract tape before position* )
    LAM symbol &$                       ( *get new symbol* )
    LAM tape LAM position
    1 #+ LAM tape LEN$ SUB$             ( *extract tape after symbol* )
    &$ ' LAM tape STO                   ( *rejoin tape and save it* )

    1 LAM position 
    LAM scanner_pos DUP
    LAM Symbol EVAL                     ( *display new symbol* )
  ;

  ABND
;

*----------------------------------------------------------------------------
* NAME: Lookup
* FUNCTION: Find next command in the list of instructions.
* STACK OPERATION: ( --> )
* CALLS: Errmsg
* CALLED BY: T_machine
*----------------------------------------------------------------------------
'
::
  0 NULL{}
  { LAM index                           ( *index to command list* )
    LAM cmnds                           ( *command list* )
  } BIND

  31 ClrUserFlag                        ( *no commmand yet* )
  LAM next_state LAM tape
  LAM position DUP SUB$ &$              ( *append symbol to state* )
  ' LAM this_state STO                  ( *save as current state* )

  "PRG" $>ID SAFE@_HERE DROP            ( *get program* )

  LAM next_state palparse DROP
  COERCE 1 #+ DUP ' LAM index STO       ( *index to command list* )
  LAM states #> ITE

    ::
      DROP                              ( *discard prog list* )
      3.1 LAM Errmsg EVAL               ( *invalid state - out of range* )
    ;

    ::
      LAM index NTHCOMDDUP              ( *get commands for current state* )
      ' LAM cmnds STO                   ( *commands for current state* )

      LENCOMP 1 #+ ONE
      DO                                ( *find relevant command* )
	LAM cmnds INDEX@ NTHCOMDDUP
	1 3 SUB$
	LAM this_state EQUAL ITE        ( *command for current state ?* )
	  ::
	    ' LAM command STO           ( *if it is, save it* )
	    31 SetUserFlag              ( *command found* )
	    ExitAtLOOP                  ( *no more searching* )
	  ;
	  DROP                          ( *if not command, lose it* )
      LOOP

      31 TestUserFlag ?SKIP

      ::
	3.3 LAM Errmsg EVAL             ( *display error message* )
      ;
    ;

  ABND
;

*----------------------------------------------------------------------------
* NAME: Setup
* FUNCTION: Initialisation.
* STACK OPERATION: ( --> )
* CALLS: Right_page_indicator, Scanner, Symbol, Frame
* CALLED BY: Main, Softkeys
*----------------------------------------------------------------------------
'
::
  LAM Valid_input EVAL                  ( *check for valid input data* )

  LAM exit ?SKIP
  ::
    LAM Frame EVAL

    THIRTEEN ZERO
    DO
      LAM space HARDBUFF
      INDEX@ 10 #* 10 GROB!             ( *display blank tape* )
    LOOP

    1 1 1 13 LAM Symbol EVAL            ( *display symbols* )

    32 ClrUserFlag                      ( *set when paging left* )
    33 ClrUserFlag                      ( *set when paging right* )

    LAM Find_last_page EVAL             ( *no of display pages* )
    LAM last_page 1 #> IT               ( *more than one page ?* )
      ::
	1 LAM Right_page_indicator      ( *display right indicator if so* )
	EVAL
      ;
    1 LAM Scanner EVAL                  ( *display scanner at start of tape* )
  ;
;

*----------------------------------------------------------------------------
* NAME: T_machine
* FUNCTION: Run the TM.
* STACK OPERATION: ( --> )
* CALLS: Direction, Lookup, Nextstate, Write
* CALLED BY: Softkeys
*----------------------------------------------------------------------------
'
::
  LAM exit ?SKIP                        ( *continue if input data is ok* )

  ::
    BEGIN

      LAM move_direction "."
      EQUAL NOT                         ( *loop until halt * )
      LAM exit NOT AND                  ( *or abort* )

    WHILE

    ::
      LAM Lookup EVAL                   ( *find command* )
      LAM move_direction "."
      EQUAL ?SKIP
	::                              ( *proceed if no errors* )
	  LAM Write EVAL                ( *write or delete symbol* )
	  LAM Direction EVAL            ( *find move direction then move* )
	  LAM move_direction "."
	  EQUAL ?SKIP
	    ::                          ( *proceed if no errors* )
	      LAM Nextstate EVAL        ( *get the next state* )
	    ;
	;
    ;

    REPEAT

* Adjust tape position pointer for correct paging when viewing result of
* TM computation. The page containing the scanner is saved for the sole
* purpose of displaying the scanner only when the page is in view.

    LAM position LAM scanner_pos #-
    1 #+ ' LAM position STO             ( *points to start of current page* )

    LAM this_page
    ' LAM scanner_page STO              ( *save page containing scanner* )
  ;

;

*----------------------------------------------------------------------------
*
{
  LAM Errmsg
  LAM Valid_input
  LAM Frame
  LAM Find_last_page
  LAM Left_page_indicator
  LAM Right_page_indicator
  LAM Symbol
  LAM Scanner
  LAM New_page
  LAM Move
  LAM Nextstate
  LAM Direction
  LAM Write
  LAM Lookup
  LAM Setup
  LAM T_machine
} BIND




